`include "cache.h"
`include "memory.h"
`include "stdbus.h"  


module Cache (
        // Asynchronous Driver Bus
        DStrobe,
        DRW,
        DAddress,
        DData,
        DReady,

        // Synchronous Memory Bus
        MStrobe,
        MRW,
        MAddress,
        MData,

        Reset,
        Clk );

input           DStrobe;
input           DRW;
input   [`ADDR] DAddress;
inout   [`DATA] DData;
output          DReady;

output          MStrobe;
output          MRW;
output  [`ADDR] MAddress;
inout   [`DATA] MData;

input           Reset;
input           Clk;

// Bidirectional buses
wire            DDataOE;        // Driver Data Output Enable
wire            MDataOE;        // Memory Data Output Enable
wire    [`DATA] MDataOut;
wire    [`DATA] DDataOut;
wire    [`DATA] DData;
wire	[`DATA]	MData;
wire    [`ADDR] MAddress;  			   
wire    [`TAG]  TagRamTag;
wire	DirtyValue, Dirty;

TagRam TagRam (
        .Address        (DAddress[`INDEX]),
        .TagIn          (DAddress[`TAG]),
        .TagOut         (TagRamTag[`TAG]),
        .Write          (Write),          
        .Clk            (Clk) );

ValidRam ValidRam (
        .Address        (DAddress[`INDEX]),
        .ValidIn        (1'b1),
        .ValidOut       (Valid),
        .Write          (Write),          
        .Reset          (Reset),
        .Clk            (Clk) );
 
 
 DirtyRam DirtyRam (
         .Address        (DAddress[`INDEX]),
         .DirtyIn        (DirtyValue),
         .DirtyOut       (Dirty),           
         .Write          (Write),      
         .Reset          (Reset),
         .Clk            (Clk) );

wire    [`DATA] DataRamDataOut;
wire    [`DATA] DataRamDataIn;
wire 	MAddrSelect;

AddrMux MaddrMux (
	.S      (MAddrSelect),
	.A      ({TagRamTag,DAddress[`INDEX],2'b0}),
	.B      (DAddress),
        .Z      (MAddress) );
        
wire	 MDataSelect;

DataMux 	MDataMux (             
        .S      (MDataSelect),      
        .A      (DataRamDataOut),             
        .B      (DData),
        .Z      (MDataOut) );      

DataMux 	DDataMux (
        .S      (DDataSelect),
        .A      (MData),
        .B      (DataRamDataOut),
        .Z      (DDataOut) );

TriStateBuffer MDataBuffer (
         .in	(MDataOut),
         .out	( MData),
         .enable(MDataOE) );
         
TriStateBuffer DDataBuffer (
         .in	(DDataOut),
         .out	(DData),
         .enable (DDataOE) );
         
         
DataRam DataRam (
        .Address        (DAddress[`INDEX]),
        .DataIn         (DData),    
        .DataOut        (DataRamDataOut),    
        .Write          (Write),
        .Clk            (Clk) );

Comparator Comparator (
        .Tag1   (DAddress[`TAG]),
        .Tag2   (TagRamTag),
        .Match  (Match) );

CacheControl CacheControl (
        .DStrobe        (DStrobe),
        .DRW            (DRW),
        .DReady         (DReady),
        .Match          (Match),
        .Valid          (Valid),
        .Dirty		(Dirty),
        .DirtyValue	(DirtyValue),
        .Write          (Write),          
        .MDataSelect    (MDataSelect), 
        .MAddrSelect	(MAddrSelect),                          
        .DDataSelect    (DDataSelect),
        .MDataOE        (MDataOE),
        .DDataOE        (DDataOE),
        .MStrobe        (MStrobe),
        .MRW            (MRW),
        .Reset          (Reset),
        .Clk            (Clk) );
endmodule
